# -----------------------------------------------------------------------------
# Programm         : C D d a t u m . p m
# -----------------------------------------------------------------------------
# Autor            : CD
# Erstellt         : 11. September 2002
# Zuletzt gendert : 26. November 2003
# Aufgabe          : Verschiedene Funktionen rund um Datum und Zeit.
# -----------------------------------------------------------------------------
#
#  $Author: usr207 $
#
#  $Date: 2003/11/26 12:50:41 $
#
#  $Id: CDdatum.pm,v 1.1 2003/11/26 12:50:41 usr207 Exp $
#
# -----------------------------------------------------------------------------
#
#  $Log: CDdatum.pm,v $
#  Revision 1.1  2003/11/26 12:50:41  usr207
#  ins CVS gestellt
#  Funktionen: heute
#              tag_der_woche
#              monatstage
#              ist_schaltjahr
#              ist_datum_zulaessig
#              ist_arbeitstag
#              ist_feiertag
#              rechnedatum
#              berechne_datum_abstand
#              berechne_ostern
#              zeit_zu_minuten
#              min_zu_hmin
#
#
# -----------------------------------------------------------------------------

package CDdatum;

#------------------------------------------------------------------------------
# Pakete / Pragmas
#------------------------------------------------------------------------------
use diagnostics;
use strict;
use warnings;

#------------------------------------------------------------------------------
# Von Exporter erben:
#------------------------------------------------------------------------------
use Exporter;
our @ISA = ('Exporter');
our @EXPORT_OK = ('heute',
                  'tag_der_woche',
                  'monatstage',
                  'ist_schaltjahr',
                  'ist_datum_zulaessig',
                  'ist_arbeitstag',
                  'ist_feiertag',
                  'rechnedatum',
                  'berechne_datum_abstand',
                  'berechne_ostern',
                  'zeit_zu_minuten',
                  'min_zu_hmin',
                 );


#------------------------------------------------------------------------------
# Globale Variablen:
#------------------------------------------------------------------------------

our @Monatsnamen = qw(nix Januar    Februar Maerz    April
                          Mai       Juni    Juli     August
                          September Oktober November Dezember);


#------------------------------------------------------------------------------
# Modulglobale Variablen:
#------------------------------------------------------------------------------

my $Debug = 0;



###############################################################################
###############################################################################
####                                                                       ####
####                 DDDD     AAA    TTTTT   U   U   M   M                 ####
####                 D   D   A   A     T     U   U   MM MM                 ####
####                 D   D   AAAAA     T     U   U   M M M                 ####
####                 D   D   A   A     T     U   U   M   M                 ####
####                 DDDD    A   A     T      UUU    M   M                 ####
####                                                                       ####
###############################################################################
###############################################################################



sub heute ($$$) {
# -----------------------------------------------------------------------------
# sub          : h e u t e
# -----------------------------------------------------------------------------
# Autor        : CD
# Aufgabe      : Gibt das Tagesdatum zurck
# Parameter    : keine
# Rckgabewert : (Tag, Monat, Jahr)
# -----------------------------------------------------------------------------
# 0.0.1 - 22.01.2003 - CD - Erstellt
# -----------------------------------------------------------------------------

    my ($t, $m, $j) = (localtime)[3,4,5];

    ++$m;       # Monate werden als 0..11 zurckgegeben
    $j += 1900; # Jahre werden als Jahre seit 1900 zurckgegeben

    print "heute: t='$t' m='$m' j='$j'\n" if $Debug;

    return ($t, $m, $j);
} # sub heute


sub tag_der_woche ($$$) {
# -----------------------------------------------------------------------------
# sub          : t a g _ d e r _ w o c h e
# -----------------------------------------------------------------------------
# Autor        : CD
# Aufgabe      : Berechnet den Wochentag zu einem Datum.
#
# Vgl. http://www.informatik.uni-ulm.de/pm/mitarbeiter/mark/day_of_week.html
#
# Parameter    : Tag (1-31), Monat (1-12) und Jahr (4-Stellig)
# Rckgabewert : 0 (Sonntag) bis 6 (Samstag)
# -----------------------------------------------------------------------------
# 0.0.1 - 17.01.2003 - CD - Erstellt
# 0.0.2 - 20.01.2003 - CD - Funktion ist_schaltjahr verwendet
# -----------------------------------------------------------------------------

    my $t = shift;
    my $m = shift;
    my $j = shift;

    print "tag_der_woche: t='$t', m='$m', j='$j'\n" if $Debug > 2;

    # Abbruch der Funktion bei falshen Argumentbergaben:
    return -1 unless defined($t) and defined($m) and defined($j);
    return -2 unless 1 <= $t and $t <= 31;
    return -3 unless 1 <= $m and $m <= 12;
    return -4 unless 1592 <= $j and $j <= 2299;

    my %jahrhundert = (15 => 0,
                       16 => 6,
                       17 => 4,
                       18 => 2,
                       19 => 0,
                       20 => 6,
                       21 => 4,
                       22 => 2,
                      );

    my $jj = $j % 100;      # letzte beiden Ziffern des Jahres
    my $jh = int($j / 100); # Jahhundert (erste beide Ziffern des Jahres

    my $schalt = ist_schaltjahr($j);
    print "jh='$jh', jj='$jj', schalt='$schalt'\n" if $Debug > 2;

    my $jhdtcode = $jahrhundert{$jh};
    print "jhdtcode='$jhdtcode'\n" if $Debug > 2;

    my $jahrcode = $jj + int($jj / 4);
    print "jahrcode='$jahrcode'\n" if $Debug > 2;

    my @monat = (1-$schalt, 4-$schalt, 4, 0, 2, 5, 0, 3, 6, 1, 4, 6);

    my $monatcode = $monat[$m-1];
    print "monatcode='$monatcode'\n" if $Debug > 2;

    my $tagcode = $jhdtcode + $jahrcode + $monatcode + $t;
    --$tagcode; # damit Sonntag auf die 0 fllt...
    $tagcode %= 7;
    print "tagcode='$tagcode'\n" if $Debug > 2;


    return $tagcode;
} # sub tag_der_woche


sub monatstage ($$) {
# -----------------------------------------------------------------------------
# sub          : m o n a t s t a g e
# -----------------------------------------------------------------------------
# Autor        : Christian Dhl
# Aufgabe      : Gibt die Anzahl der Tage des angegebenen Monats/Jahres an
# Parameter    : Monat, Jahr (fr Schaltjahre)
# Rckgabewert : Anzahl Tage (28 bis 31)
# -----------------------------------------------------------------------------
# 0.0.1 - 07.08.2003 - CD - Erstellt
# -----------------------------------------------------------------------------

    my $monat = shift;
    my $jahr  = shift;

    die "Fehler!" unless defined($monat) and defined($jahr);

    my @l = (31,
             28+ist_schaltjahr($jahr),
             31,
             30,
             31,
             30,
             31,
             31,
             30,
             31,
             30,
             31);

    return $l[$monat-1];
} # sub monatstage


sub ist_schaltjahr ($) {
# -----------------------------------------------------------------------------
# sub          : i s t _ s c h a l t j a h r
# -----------------------------------------------------------------------------
# Autor        : CD
# Aufgabe      : Gibt an, ob das bergebene Jahr ein Schaltjahr ist.
# Parameter    : Jahr
# Rckgabewert : 0 oder 1
# -----------------------------------------------------------------------------
# 0.0.1 - 20.01.2003 - CD - Erstellt
# -----------------------------------------------------------------------------

    my $j = shift;

    return 1 if $j%4==0 and ($j%100!=0 or $j%400==0);
    return 0;
} # sub ist_schaltjahr


sub ist_datum_zulaessig ($$$) {
# -----------------------------------------------------------------------------
# sub          : i s t _ d a t u m _ z u l a e s s i g
# -----------------------------------------------------------------------------
# Autor        : CD
# Aufgabe      : Testet, ob ein Datum zulssig ist.
# Parameter    : Tag, Monat, Jahr
# Rckgabewert : 0 = unzulssig, 1 = zulssig
# -----------------------------------------------------------------------------
# 0.0.1 - 15.01.2003 - CD - Erstellt
# 0.0.2 - 20.01.2003 - CD - Array @l und ist_schaltjahr() verwendet.
# -----------------------------------------------------------------------------

    my $t = shift;
    my $m = shift;
    my $j = shift;

    return 0 unless defined($t) and defined($m) and defined($j);

    my @l = (31, 28+ist_schaltjahr($j), 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);
    return 0 if ($t < 1 or $l[$m-1] < $t) or ($m < 1 or 12 < $m);
    return 1;
} # sub ist_datum_zulaessig


sub ist_arbeitstag ($$$) {
# -----------------------------------------------------------------------------
# sub          : i s t _ a r b e i t s t a g
# -----------------------------------------------------------------------------
# Autor        : CD
# Aufgabe      : Es wird berechnet, ob der Tag ein Arbeitstag ist.
# Parameter    : Tag, Monat, Jahr
# Rckgabewert : 0 oder 1
# -----------------------------------------------------------------------------
# 0.0.1 - 20.01.2003 - CD - Erstellt
# -----------------------------------------------------------------------------

    my $t = shift;
    my $m = shift;
    my $j = shift;


    return 0 unless defined($t) and defined($m) and defined($j);

    my $wt = tag_der_woche($t, $m, $j);


    return 0 if $wt == 0 or $wt == 6; # Samstage und Sonntage
    return 0 if ist_feiertag($t, $m, $j);


    return 1;
} # sub ist_arbeitstag


sub ist_feiertag ($$$) {
# -----------------------------------------------------------------------------
# sub          : i s t _ f e i e r t a g
# -----------------------------------------------------------------------------
# Autor        : CD
# Aufgabe      : Berechnet, ob der angegebene Tag ein Feiertag ist.
# Parameter    : Tag, Monat, Jahr
# Rckgabewert : 0 oder 1
# -----------------------------------------------------------------------------
# 0.0.1 - 20.01.2003 - CD - Erstellt
# -----------------------------------------------------------------------------

    my $t = shift;
    my $m = shift;
    my $j = shift;

    return 0 unless defined($t) and defined($m) and defined($j);

    my %feiertage_var = (-2 => 'Karfreitag',
                          1 => 'Ostermontag',
                         39 => 'Christi Himmelfahrt',
                         50 => 'Pfingstmontag',
                         60 => 'Fronleichnam'
                        );
    my %feiertage_fix = ('01.01.' => "Neujahr",
                         '06.01.' => "Dreikoenigstag",
                         '01.05.' => "Tag der Arbeit",
                         '19.06.' => 'Fronleichnam',
                         '15.08.' => "Mariae Himmelfahrt",
                         '03.10.' => 'Tag der deutschen Einheit',
                         '01.11.' => 'Allerheiligen',
                         '25.12.' => '1. Weihnachtstag',
                         '26.12.' => '2. Weihnachtstag',
                        );

    # fixe Feiertage:
    my $f = sprintf("%02d.%02d.", $t, $m);
    if (defined($feiertage_fix{$f})) {
        print "Feiertag '", $feiertage_fix{$f}, "' gefunden!\n" if $Debug > 1;
        return 1;
    }

    #variable Feiertage:
    my ($ot, $om, $oj) = berechne_ostern($j); # Ostersonntag berechnen
    printf "Ostersonntag ist der %02d.%02d.%4d.\n", $ot, $om, $oj
        if $Debug > 1;
    for my $k (keys %feiertage_var) {
        my ($tt, $tm, $tj) = rechnedatum($ot, $om, $oj, $k);
        printf "%-20s : %02d.%02d.%4d\n", $feiertage_var{$k}, $tt, $tm, $tj
            if $Debug > 1;
        return 1 if $t==$tt and $m==$tm and $j=$tj;
    }
} # sub ist_feiertag


sub rechnedatum ($$$$) {
# -----------------------------------------------------------------------------
# sub          : r e c h n e d a t u m
# -----------------------------------------------------------------------------
# Autor        : CD
# Aufgabe      : Addiert (oder subtrahiert bei negativen Zahlen) die angegebene
#                Anzahl Tage zum Datum.
# Parameter    : Tag, Monat, Jahr, Anzahl-Tage
# Rckgabewert : (Tag, Monat, Jahr)
# -----------------------------------------------------------------------------
# 0.0.1 - 20.01.2003 - CD - Erstellt
# -----------------------------------------------------------------------------

    my $t     = shift;
    my $m     = shift;
    my $j     = shift;
    my $delta = shift;

    die "falsche Parameter in Funktion CDdatum::rechnedatum!"
        unless scalar(@_) == 0 and
               defined($t)     and
               defined($m)     and
               defined($j)     and
               defined($delta)   ;

    my @l = (31, 28+ist_schaltjahr($j), 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);

    printf "Startdatum: %02d.%02d.%4d + %d Tage\n", $t, $m, $j, $delta
        if $Debug > 1;
    $t += $delta;

    # negatives Delta:
    while ($t <= 0) {
        --$m;
        if ($m == 0) {
            $m = 12;
            --$j;
        }
        $t += $l[$m-1];
    }
    # 00.05.2003 -> 30.04.2003
    # -3.05.2003 -> 27.04.2003

    # positives Delta:
    while ($t > $l[$m-1]) {
        $t -= $l[$m-1];
        ++$m;
        if ($m == 13) {
            $m = 1;
            ++$j;
        }
    }
    # 32.05.2003 -> 01.06.2003

    printf "Zieldatum:  %02d.%02d.%4d\n", $t, $m, $j if $Debug > 1;
    return ($t, $m, $j);
} # sub rechnedatum


sub berechne_datum_abstand ($$$$$$) {
# -----------------------------------------------------------------------------
# sub          : b e r e c h n e _ d a t u m _ a b s t a n d
# -----------------------------------------------------------------------------
# Autor        : Christian Dhl
# Aufgabe      : Berechnet den Abstand zweier Dati in Tagen
# Parameter    : Tag1, Monat1, Jahr1, Tag2, Monat2, Jahr2
# Rckgabewert : Anzahl Tage zwischen zweitem und erstem Datum (Datum1-Datum2)
# -----------------------------------------------------------------------------
# 0.0.1 - 07.08.2003 - CD - Erstellt
# -----------------------------------------------------------------------------

    my $tt = shift;
    my $mm = shift;
    my $jj = shift;
    my $t  = shift;
    my $m  = shift;
    my $j  = shift;

    my $abstand = 0;


    die "falsche Parameter in Funktion CDdatum::rechnedatum!"
        unless scalar(@_) == 0 and
        defined($tt)           and
        defined($mm)           and
        defined($jj)           and
        defined($t )           and
        defined($m )           and
        defined($j )             ;


    #print "-"x77,"\n";
    #printf "%02d.%02d.%04d - %02d.%02d.%04d\n", $tt, $mm, $jj, $t, $m, $j;

    #--------------------------------------------------------------------------
    # Tagesdifferenz:
    #--------------------------------------------------------------------------
    # Tage gleich machen, Abstand merken
    $abstand = $tt - $t;
    $tt = $t;
    # 31.01.2000 29.02.2000 -> beide Tage auf 29, Abstand ist 2
    # 29.02.2000 31.03.2000 -> beide Tage auf 31, Abstand ist -2
    #print "Tage gleichgemacht, Abstand ist nun '$abstand'\n";
    #printf "%02d.%02d.%04d - %02d.%02d.%04d\n", $tt, $mm, $jj, $t, $m, $j;

    #--------------------------------------------------------------------------
    # Monatsdifferenz:
    #--------------------------------------------------------------------------
    while ($mm > $m) {
        $abstand += monatstage($m, $j);
        ++$m;
    }
    while ($mm < $m) {
        $abstand -= monatstage($mm, $jj);
        ++$mm;
    }
    #print "Monate gleichgemacht, Abstand ist nun '$abstand'\n";
    #printf "%02d.%02d.%04d - %02d.%02d.%04d\n", $tt, $mm, $jj, $t, $m, $j;

    #--------------------------------------------------------------------------
    # Jahresdifferenz:
    #--------------------------------------------------------------------------
    while ($jj > $j) {
        if (ist_schaltjahr($j)   and $mm <  3 or
            ist_schaltjahr($j+1) and $mm >= 3)
        {
            $abstand += 366;
        }
        else {
            $abstand += 365;
        }
        ++$j;
    }
    while ($jj < $j) {
        if (ist_schaltjahr($jj)   and $mm <  3 or
            ist_schaltjahr($jj+1) and $mm >= 3)
        {
            $abstand -= 366;
        }
        else {
            $abstand -= 365;
        }
        ++$jj;
    }
    #print "Jahre gleichgemacht, Abstand ist nun '$abstand'\n";
    #printf "%02d.%02d.%04d - %02d.%02d.%04d\n\n", $tt, $mm, $jj, $t, $m, $j;

    return $abstand;
} # sub berechne_datum_abstand


sub berechne_ostern ($) {
# -----------------------------------------------------------------------------
# sub          : b e r e c h n e _ o s t e r n
# -----------------------------------------------------------------------------
# Autor        : CD
# Aufgabe      : Berechnet Ostern nach Gausschem Algorithmus
#
# vgl. http://home.nikocity.de/geelhaar/spielwiese/ostern/ostern.html
#
# Parameter    : Jahr
# Rckgabewert : Array (Tag, Monat, Jahr)
# -----------------------------------------------------------------------------
# 0.0.1 - 20.01.2003 - CD - Erstellt
# -----------------------------------------------------------------------------

    my $j     = shift;       # Jahr (Eingabe und Ausgabe)

    die "falsche Parameter in Funktion CDdatum::berechne_ostern!"
        unless scalar(@_) == 0 and defined $j;

    my $var_a = $j % 19;     # Variable des Algorithmus
    my $var_b = $j %  4;     # Variable des Algorithmus
    my $var_c = $j %  7;     # Variable des Algorithmus
    my $var_d = 0;           # Variable des Algorithmus
    my $var_e = 0;           # Variable des Algorithmus
    my $var_m = 0;           # Variable des Algorithmus, wird aus %m besetzt
    my $var_n = 0;           # Variable des Algorithmus, wird aus %n besetzt
    my $jh    = int($j/100); # Jahrhundert
    my $t     = 0;           # Tag (Ausgabe)
    my $m     = 0;           # Monat (Ausgabe)


    if (17 <= $jh and $jh < 22) {
        my %m  = (17 => 23, 18 => 23, 19 => 24, 20 => 24, 21 => 24);
        my %n  = (17 =>  3, 18 =>  4, 19 =>  5, 20 =>  5, 21 =>  6);
        $var_m = $m{$jh};
        $var_n = $n{$jh};
    }
    else {
        die "Kann das Osterdatum nur zwischen 1700 und 2199 berechnen.\n";
    }

    $var_d = (19 * $var_a + $var_m) % 30;
    $var_e = (2 * $var_b + 4 * $var_c + 6 * $var_d + $var_n) % 7;

    $t = $var_d + $var_e + 22;
    $m = 3;

    if ($t > 31) {
        $t = $var_d + $var_e - 9;
           $m = 4;
    }

    $t = 19 if $m == 4 and $t == 26;
    $t = 18 if $m == 4 and $t == 25 and $var_d == 28 and $var_a > 10;


    return ($t, $m, $j);
} # sub berechne_ostern





###############################################################################
###############################################################################
####                                                                       ####
####                     ZZZZZ   EEEEE     I     TTTTT                     ####
####                        Z    E         I       T                       ####
####                       Z     EEEE      I       T                       ####
####                      Z      E         I       T                       ####
####                     ZZZZZ   EEEEE     I       T                       ####
####                                                                       ####
###############################################################################
###############################################################################



sub zeit_zu_minuten ($) {
# -----------------------------------------------------------------------------
# sub          : z e i t _ z u _ m i n u t e n
# -----------------------------------------------------------------------------
# Autor        : CD
# Aufgabe      : Macht aus einer Zeit der Form 9:30 Minuten seit Mitternacht.
# Parameter    : Zeit
# Rckgabewert : Anzahl Minuten
# -----------------------------------------------------------------------------
# 0.0.1 - 15.01.2003 - CD - Erstellt
# -----------------------------------------------------------------------------

    my ($h, $m) = split(/:/, $_[0], 2);
    return 0 unless defined($h) and defined($m);
    return 60*$h + $m;
} # sub zeit_zu_minuten


sub min_zu_hmin ($) {
# -----------------------------------------------------------------------------
# sub          : m i n _ z u _ h m i n
# -----------------------------------------------------------------------------
# Autor        : CD
# Aufgabe      : Wandelt Minuten in Stunden:Minuten (150 -> 2:30)
# Parameter    : Minuten
# Rckgabewert : Stunden:Minuten
# -----------------------------------------------------------------------------
# 0.0.1 - 15.01.2003 - CD - Erstellt
# -----------------------------------------------------------------------------

    my $min  =  $_[0];
    my $vorz =  ($min<0?"-":"");
    $min     *= -1 if $min < 0;
    my $h    =  int($min/60);
    my $m    =  $min - $h*60;
    my $r    =  "$vorz$h:" . ($m<10?"0":"") . "$m";
    return $r;
} # sub min_zu_hmin



return 1;



